/*jslint undef: true, widget: true, browser: true, onelet: true, eqeqeq: true */
/*global YAHOO, iw_js_data, IW_AC_DATA, iw_xhr, iw_form_validate_handler, updateStatusClass, iw_lib, iw_ctrl_handler, iw_form_handler */

var iw_form_autocomplete_data_handler;
var iw_form_generate_password_data_handler;
var iw_form_label_handler;

/*
 * Main iw_form object.  Everything related to interworx forms and javascript goes here.
 */
var iw_form = {
    /*
     * Initialize the iw_form object and any interworx forms present on the page.
     */
    init: function () {
        let shells = $('.iw-form-shell').get();

        if (iw_js_data.iw_env && iw_js_data.iw_env.is_node === true) {
            this.initLocalChangesForms();
        }

        $(document).on('click', '.iw-link-action', this.executeActionLink);

        this.initUnlimited();
        this.initSubmits(shells);
        this.initEventHandlers(shells);
        this.initShowSelect();
        this.initToolbarActions();
    },

    /*
     * Popups up the "Loading..." popup, before full form load.
     */
    showLoadingPopup: function () {
        iw_lib.popupRightModal(
            '',
            '<i class="' + IW.icons._BASE + ' ' + IW.icons.PENDING + ' ' + IW.icons.SPIN + '"> ##LG_LOADING##'
        );
    },

    /*
     * A "handled" action link is clicked.
     *
     * Take over the click and xhr request the action.
     *
     * @param Event e
     */
    executeActionLink: function (e) {
        let ctrl, action, input, callback, href;
        e.stopPropagation();
        e.preventDefault();

        if ($(this).hasClass('iw-link-disabled')) {
            return;
        }

        if (!$(this).data('iw-no-popup')) {
            iw_form.showLoadingPopup();
        }

        if ($(this).data('iw-href')) {
            href = $(this).data('iw-href');
        } else {
            href = this.href;
        }

        callback = function (reply) {
            let htmlResponse = reply.getReplyData().html;
            if (htmlResponse && htmlResponse.includes('iw-form-shell')) {
                // will execute once form is fully loaded; same as on load
                $(function () {
                    let shell_id = $(htmlResponse).first('div').attr('id');
                    iw_form.initAutoCompleteShell($(`#${shell_id}`));
                })
            }
        };

        if ($(this).data('iw-refresh')) {
            callback = function () {
                location.reload();
            };
        }

        ctrl = iw_form.getCtrlFromUrl(href);
        action = iw_form.getActionFromUrl(href);
        input = iw_form.getInputFromUrl(href);

        iw_ctrl_handler.execute(null, callback, ctrl, action, input);
    },

    /*
     * Given a URL, get the action=blah part from it.
     *
     * @param string url
     * @return string
     */
    getActionFromUrl: function (url) {
        let data = this.getQueryStringDataFromUrl(url);

        if (data.action) {
            return data.action;
        }
        return '';
    },

    /*
     * Get the query string data from the url (excluding the action).
     *
     * @param string url
     * @return array
     */
    getInputFromUrl: function (url) {
        let input = this.getQueryStringDataFromUrl(url);

        if (input.action) {
            delete input.action;
        }

        return input;
    },

    /*
     * Get the query string data from url url.
     *
     * @param string url
     * @return array
     */
    getQueryStringDataFromUrl: function (url) {
        let parts, pairs, pair, temp, output = {};

        parts = url.split('?');

        if (parts[1]) {
            pairs = parts[1].split('&');

            for (pair in pairs) {
                if (pairs.hasOwnProperty(pair)) {
                    pair = pairs[pair];
                    temp = pair.split('=');
                    output[temp[0]] = temp[1];
                }
            }
        }

        return output;
    },

    /*
     * Given a url, determine the interworx controller portion.
     *
     * @param string url
     * @return string
     */
    getCtrlFromUrl: function (url) {
        let parts, startPosition, endPosition;
        if (url.substring(0, 1) === '/') {
            parts = url.split('?');
            return parts[0];
        } else if (url.substring(0, 4).toLowerCase() === 'http') {
            startPosition = url.indexOf('worx/') - 5;
            endPosition = url.indexOf('?') || url.length;
            return url.substring(startPosition, endPosition);
        }
    },

    getSubmitButton: function (form) {
        let name   = $(form).attr('name'),
            button = document.getElementById(name + '-submit');

        if (!button) {
            button = $(form).find('#submit').get()[0];
        }

        return button;
    },

    setupLabelRefresh: function (form) {
        let input = {},
            _refresh;

        _refresh = function () {
            input.form = form.iw_form_class_name.value;
            input.ctrl = form.iw_ctrl_name.value;
            iw_xhr.getReply(iw_form_label_handler, input, form);
        };

        iw_xhr.subscribeTo(_refresh, 'addCommit');
        iw_xhr.subscribeTo(_refresh, 'deleteCommit');
    },

    /*
     * Hide a group.
     *
     * @param string group
     */
    _hideGroup: function (group) {
        $(this._getGroupElements(group)).css('display', 'none');
    },

    /*
     * Show a group.
     *
     * @param string group
     */
    _showGroup: function (group) {
        $(this._getGroupElements(group)).css('display', '');
    },

    /*
     * Get elements in a group.
     *
     * @param string group
     * @return array
     */
    _getGroupElements: function (group) {
        return $('tr[data-iw-shown=' + group + ']');
    },

    /*
     * Toggle a show select.
     *
     * @param Event e
     */
    toggleShowSelect: function (e) {
        let option     = this.options[this.selectedIndex],
            opt_length = this.options.length,
            group, opt_group, i;

        group = $(option).data('iw-show');

        if (group) {
            iw_form._showGroup(group);
        }

        for (i = 0; i < opt_length; ++i) {
            opt_group = $(this.options[i]).data('iw-show');

            if (opt_group && opt_group !== group) {
                iw_form._hideGroup(opt_group);
            }
        }
    },

    /*
     * Toggle show...?
     *
     * @param Event e
     */
    toggleShow: function (e) {
        let group, form, others, other_group, others_length, i;

        group = $(this).data('iw-show');

        if (group && (this.checked === true || this.selected === true)) {
            iw_form._showGroup(group);
        } else {
            iw_form._hideGroup(group);
        }

        form = $(this).closest('form');

        if (this.tagName.toUpperCase() === 'INPUT' && this.type === 'radio') {
            others = $('.in-' + this.name, form);
            others_length = others.length;

            for (i = 0; i < others_length; ++i) {
                if (others[i].id !== this.id) {
                    other_group = $(others[i]).data('iw-show');

                    if (other_group) {
                        iw_form._hideGroup(other_group);
                    }
                }
            }
        }
    },

    /*
     * Toggle show flag.
     *
     * @param Event e
     */
    toggleShowFlag: function (e) {
        let elements = $('.iw-show-' + this.id),
            display;

        if (this.checked === true) {
            display = '';
        } else {
            display = 'none';
        }

        $(elements).css('display', display);
    },

    /*
     * Init local changes on forms.
     */
    initLocalChangesForms: function () {
        let checkboxes     = $('.in-enable_local_changes'),
            checkbox_count = checkboxes.length,
            form, i;

        for (i = 0; i < checkbox_count; i++) {
            form = $(checkboxes[i]).closest('form');

            this.disableForm(form, ['enable_local_changes']);
        }
    },

    /*
     * On enable local changes
     *
     * @param Event e
     */
    onEnableLocalChanges: function (e) {
        let form = $(this).closest('form');
        iw_form.enableForm(form);
    },

    disableForm: function (form, exceptions) {
        let inputs      = $('.iw-in', form),
            input_count = inputs.length,
            i;
        exceptions = exceptions || [];

        for (i = 0; i < input_count; ++i) {
            if (exceptions.indexOf(inputs[i].name) === -1) {
                inputs[i].disabled = 1;
            }
        }

        iw_form.getSubmitButton(form).disabled = 1;
    },

    /*
     * Enable form.
     *
     * @param HTMLElement form
     */
    enableForm: function (form) {
        let inputs      = $('.iw-in', form),
            input_count = inputs.length,
            i;

        for (i = 0; i < input_count; ++i) {
            inputs[i].disabled = 0;
        }

        iw_form.getSubmitButton(form).disabled = 0;
    },

    formSubmitStatusCache: {},

    /*
     * Switch between file upload, and choose a local file.
     *
     * @param Event e
     */
    fileSwitch: function (e) {
        let up_lang        = '##LG_SWITCH_TO_FILE_UPLOAD##',
            local_lang     = '##LG_SWITCH_TO_FILE_LOCAL##',
            len            = this.id.length,
            id             = this.id.substr(0, len - 9),
            input          = document.getElementById(id),
            input_name     = input.name,
            alt_input      = document.getElementById(id + '_alt');
        e.stopPropagation();
        e.preventDefault();

        input.name = alt_input.name;
        input.id = alt_input.id;
        input.value = '';

        alt_input.name = input_name;
        alt_input.id = id;
        let place_holder = $(this).prev();
        $(place_holder).addClass('iw-local-file');
        $(input).css('display', 'none');
        $(this).css('display', '');
        $(alt_input).css('display', '');
    },

    /*
     * Initialize the auto-complete logic.
     *
     * Note - this needs to be invoked every time a form is
     * added to the DOM - e.g. add/edit actions
     */
    initAutoCompleteShell: function (shell) {
        $(shell).find('.iw-form-ac').each(function (index, input) {
            iw_form.initAutoCompleteInput(input);
        })
    },

    /*
     * Initialize the auto-complete logic on individual input field.
     */
    initAutoCompleteInput: function (input) {
        let form = $(input).closest('form');

        /*
         * The awesome bar needs to be instantiated differently, done elsewhere
         */
        if (input.id === 'iw_awesome_input') {
            return;
        }

        if (input.value === '') {
            $(input).attr('placeholder', '##LG_BEGIN_TYPING_FOR_SUGGESTIONS##');
        }

        /*
         * If the autocomplete data is missing, fetch it from the server.
         *
         * The callback that does the fetching will call this function again, which will
         * do the rest of the initialization.  We filled in the "begin typing" to look
         * like it's ready.
         */
        if (IW_AC_DATA.hasOwnProperty(input.id) === false) {
            let args = iw_form.getHiddenInputValues(form);

            args.input = $(input).attr('data-ac-name') || input.name;
            args.form = $('.in-iw_form_class_name', form).val();

            iw_xhr.getReply(iw_form_autocomplete_data_handler, args, input);
            return;
        }

        let ac_data = new Bloodhound(
            {
                datumTokenizer: Bloodhound.tokenizers.obj.whitespace('d'),
                queryTokenizer: Bloodhound.tokenizers.whitespace,
                identify: function (obj) {
                    return obj.d;
                },
                local: IW_AC_DATA[input.id]
            }
        );

        let AC = $('#' + input.id);
        AC.typeahead(
            null,
            {
                source: ac_data,
                display: 'd',
                limit: 10
            }
        );
    },

    /*
     * On Auto-complete expand.
     *
     * @param string? sType
     * @param array   args
     */
    onACExpand: function (sType, args) {
        let AC        = args[0],
            input     = AC.getInputEl(),
            height    = input.clientHeight,
            container = AC.getContainerEl();
        $(container).css('top', height + 'px');
    },

    isSetAutoComplete: function (input) {
        var td          = $(input).closest('td'),
            checks        = $(td).find('.iw-form-ac-check');

        return checks.length > 0;
    },

    registeredACs: {},

    onACFocus: function () {
        if (this.value === '##LG_BEGIN_TYPING_FOR_SUGGESTIONS##') {
            this.value = '';
        }
        $(this).removeClass('iw-dim');
    },

    onACSelect: function (sType, args) {
        let AC      = args[0],
            data    = args[2],
            input   = AC.getInputEl(),
            old_val = input.value.split("\n"),
            range;
        iw_form.clearStatus(input);

        // Remove the partial match
        old_val.pop();

        old_val.push(data.v);

        input.value = old_val.join("\n") + "\n";

        // Move the caret for IE
        if (input.createTextRange) {
            range = input.createTextRange();
            range.collapse(false);
            range.select();
        }
    },

    isInPopup: function (el) {
        let modal = $(el).closest('#iw-right-modal').length;

        return (modal);
    },

    onAutoCompleteClick: function (e) {
        let label = $(e.target).text();
        let inputValue = label.substring(0, label.indexOf('@'));
        let span = $(e.target).parents('span')[0];
        let first = $(span).children()[0];
        let acName = $(first).attr('data-ac-name')
        let inputId = `iw-in-ac-${acName}-${inputValue}`;

        let tableSelector = `#iw-in-${acName}_j_table`;
        let labelSelector = `label:contains("${label}")`;

        // can't search by id - id is different for preexisting groups, e.g. edit form
        let notSelected = $(tableSelector).find(labelSelector).length === 0;

        if (notSelected) {
            iw_form.addSelectedAutocompleteItem(inputId, acName, inputValue, label);
        }

        // clear suggestions input after selection
        let input_suggestions = $(`#iw-in-${acName}_ac_j`).get()[0];
        input_suggestions.value = '';
    },

    addSelectedAutocompleteItem: function (id, name, value, label) {
        let inputElement = `
            <input
                id="${id}"
                name="${name}[]"
                value="${value}"
                class="iw-in in-${name}"
                type="checkbox"
                checked="checked"
                data-iw-id="iw-in-${name}_j"
            >
        `;

        let labelElement = `<label for ="${id}">${label}</label>`;

        let $tr = $(`
            <tr class="iw-form-unsaved">
                <td>${inputElement}</td>
                <td>${labelElement}</td>
            </tr>`);

        $(`#iw-in-${name}_j_table tbody`).append($tr);
    },

    /*
     * @return el|boolean
     */
    valueAlreadyExists: function (value, table) {
        let inputs        = $(table).children('input'),
            inputs_length = inputs.length,
            i;

        if (!inputs) {
            return false;
        }

        for (i = 0; i < inputs_length; ++i) {
            if (inputs[i].value === value) {
                return inputs[i];
            }
        }

        return false;
    },

    getInputNameFromACName: function (ac_input) {
        // Whacks off the "_ac"
        return ac_input.name.substring(0, ac_input.name.length - 3);
    },

    focusCustomRadio: function (e) {
        let div   = $(this).closest('div'),
            radio = $('.iw-form-radio-custom', div);

        radio[0].checked = true;
    },

    submitEvents: {},

    getFormActions: function (form) {
        let input  = form.iw_action_name,
            options,
            option_count,
            output = [],
            i;

        if (!input) {
            return [];
        }

        if (input.tagName.toLowerCase() === 'input') {
            return [input.value];
        }

        if (input.tagName.toLowerCase() === 'select') {
            options = $(input).children('option');
            option_count = options.length;

            for (i = 0; i < option_count; ++i) {
                output.push(options[i].value);
            }
        }

        return output;
    },

    getFormNameFromShell: function (shell) {
        return $(shell).attr('id').substring(0, $(shell).attr('id').length - 6);
    },

    getFormFromShell: function (shell) {
        let form_name = iw_form.getFormNameFromShell(shell);

        return document.forms[form_name];
    },

    initSubmits: function (shells) {
        let form, form_name, shell;

        $.each(shells, function (index, shell) {
                form_name = iw_form.getFormNameFromShell(shell);
                form = iw_form.getFormFromShell(shell);

                iw_form.formSubmitStatusCache[form_name] = false;

            if (iw_form.isHandled(form)) {
                iw_form_handler(form, null, 'submit');
            }

            if ($(form).hasClass('iw-form-label-refresh')) {
                iw_form.setupLabelRefresh(form);
            }
        });//end for
    },

    initToolbarActions: function () {
        $(document).off('click', '.btn-toolbar-action');

        $('.btn-toolbar-action').each((index, element) => {
            let form = $(element).closest('form').get()[0];
            iw_form_handler(form, element, 'click');
        })
    },

    isHandled: function (form) {
        if (form.no_handler) {
            return false;
        }
        return true;
    },

    tooltip_instances: [],

    /*
     * Initializes unlimited checkboxes
     */
    initUnlimited: function () {
        let unlimChecks = $('.iw-unlim-check'),
            unlimChecks_length = unlimChecks.length,
            i, div, input;

        for (i = 0; i < unlimChecks_length; ++i) {
            div = $(unlimChecks[i]).closest('div');
            input = $(div).children('input')[0];
            if (input.value === '999999999') {
                unlimChecks[i].checked = 1;
                iw_form.doUnlimitedCheck(unlimChecks[i]);
            } else {
                unlimChecks[i].checked = 0;
                iw_form.doUnlimitedCheck(unlimChecks[i]);
            }
        }
    },

    /*
     * Init show-on selects.
     */
    initShowSelect: function () {
        let selects = $('select.iw-show'),
            select,
            option,
            opt_length,
            group,
            opt_group,
            i;

        $.each(selects, function (select) {
            if (selects.hasOwnProperty(select)) {
                select = selects[select];
                option = select.options[select.selectedIndex];
                group = option.getAttribute('data-iw-show');

                if (group) {
                    iw_form._showGroup(group);
                }
                opt_length = select.options.length;
                for (i = 0; i < opt_length; ++i) {
                    opt_group = select.options[i].getAttribute('data-iw-show');
                    if (opt_group && opt_group !== group) {
                        iw_form._hideGroup(opt_group);
                    }
                }
            }
        });
    },

    refresh: function (form) {
        let f = $(form);

        let _callback = function (reply, options) {
            // Prevent the handle208 from doing it's usual thing
            options.prevent_default = true;

            let shell = f.closest('.iw-form-shell');
            $(shell).html(reply.getReplyData().html);
            iw_form.initSubmits([shell]);
        };

        let _input = {
            ctrl          : iw_form.getCtrl(f),
            action        : iw_form.getAction(f),
            form          : $('input[name=iw_form_class_name]', f).val(),
            security_token: iw_js_data.iw_env.security_token
        };

        iw_ctrl_handler.execute(null, _callback, 'index', 'refreshFormCommit', _input);
    },

    /*
     * Sets up form-related event handlers that can be delegated
     *
     * @param array   shells
     */
    initEventHandlers: function (shells) {
        let forms;


        // Submit handlers
        $.each(shells, function (index, shell) {
            iw_form.initAutoCompleteShell(shell);

            forms = $('.iw-form', shell).get();
            iw_form.initFormEvents(forms);
        });


        $(document).on('click', '.iw-form-shell .tt-suggestion', this.onAutoCompleteClick);
        $(document).on('focus', '.iw-form-ac', this.onACFocus);

        // TIMDO: goes away with new unlimiter options.
        $(document).on('click', '.iw-unlimiter', this.unlimitedCheck);
        $(document).on('click', 'input[data-iw-show]', this.toggleShow);
        $(document).on('click', '.iw-password-gen', this.generatePassword);

        $(document).on('click', 'a.iw-check-all', this.checkAll);
        $(document).on('click', 'a.iw-check-none', this.checkNone);

        $(document).on('click', 'input.iw-in[type=checkbox]', this._clearStatus);
        $(document).on('click', '.iw-file-switch', this.fileSwitch);

        $(document).on('click', 'input.iw-form-cancel', this.onCancel);

        $(document).on('click', 'input-in-enable_local_changes', this.onEnableLocalChanges);

        $(document).on('focusin', '.iw-in', this._clearStatus);
        $(document).on('focusout', '.iw-in', this.validate);

        $(document).on('dblclick', 'tr.iw-form-group-hd', this.toggleGroupDisplay);
        $(document).on('click', 'tr.iw-form-group-hd .iw-collapse', this.toggleGroupDisplay);

        iw_form.initGroupStates();

        $(document).on('submit', 'form', this.onSubmit);

        $(document).on('change', 'select.iw-show', this.toggleShowSelect);
    },

    toggleGroupDisplay(e) {
        let target = $(e.target).closest('tr'),
            group_num = $(target).data('iw-group-num'),
            status = $(target).find('.iw-collapse');

        e.preventDefault();
        e.stopPropagation();

        $('tr.group-num-' + group_num).toggleClass('iw-hide-group');
        status.toggleClass(IW.icons.PLUS + ' ' + IW.icons.MINUS);
    },

    initGroupStates() {
        $('tr.iw-form-group-hd').each(
            function () {
                let group_num = $(this).data('iw-group-num'),
                    status = $(this).find('.iw-collapse');

                if (status.hasClass(IW.icons.PLUS)) {
                    $('tr.group-num-' + group_num).addClass('iw-hide-group');
                }
            }
        )
    },

    /*
     * Connects the events for the form.
     *
     * These are the events that cannot be delegated, due to browser incompatibilities
     * or other inconveniences.
     *
     * TIMDO: Bootstrap/jQuery can delegate at least some of these now.
     *
     * @param string|HTMLElement|array form_ish
     */
    initFormEvents: function (form_ish) {
        let forms, form, action, text_collapse, textarea, textdiv;

        if (typeof form_ish == 'string') {
            form_ish = $('#' + form_ish);
        }

        if ($.isArray(form_ish) === false) {
            forms = [form_ish];
        } else {
            forms = form_ish;
        }

        for (form in forms) {
            if (forms.hasOwnProperty(form)) {
                form = forms[form];


                text_collapse = $('.iw-form-text-collapse', form).get();

                for (textarea in text_collapse) {
                    if (text_collapse.hasOwnProperty(textarea)) {
                        textarea = text_collapse[textarea];

                        textdiv = textarea.closest('.iw-form-input');
                        if ($(textdiv).hasClass('iw-mult') === false) {
                            this.initTextarea(textarea);
                        }
                    }
                }

                $('.iw-radio-custom-input', form).on('focus', this.focusCustomRadio);

                $(text_collapse).focus(this.listenForNewlines);
                $(text_collapse).blur(this.removeListenForNewlines);
                $(text_collapse).keyup(this.updateCharCount);

                $('.in-password-with-generator', form).keyup(this.keydownPasswordStrength);
                $('.in-password-with-generator', form).bind('validateCompleteEvent', this.updateStrengthMessage);

                action = this.getAction(form);

                // Anything that appears to be an add*Commit action gets the reset enabled
                if (action.match(/^add[a-zA-Z]*Commit$/)) {
                    iw_xhr.subscribeTo(
                        this.enableResetOption,
                        action,
                        this.getCtrl(form)
                    );
                }
            }//end if
        }//end for
    },

    deleteOptionByValue: function (select, value) {
        let length = select.length, i;
        for (i = 0; i < length; i++) {
            if (select[i].value === value) {
                select.remove(i);
                length = select.length;
            }
        }
    },

    newlineKeyListener: false,
    keyListenerTarget : false,

    adjustCollapse: function (e) {
        let el          = e.target,
            value       = $(el).val(),
            length      = value.split("\n").length,
            currentRows = Number($(el).attr('rows'));

        if (length >= currentRows && currentRows < $(el).attr('data-maxrows')) {
            $(el).attr('rows', currentRows + 1);
        }
    },

    listenForNewlines: function (e) {
        $(this).keypress(function (e) {
            if (e.which === 10 || e.which === 13) {
                iw_form.adjustCollapse(e);
            }
        })
    },

    removeListenForNewlines: function (e) {
        if (iw_form.newlineKeyListener) {
            iw_form.newlineKeyListener.disable();
        }
    },

    initTextarea: function (input) {
        let limit   = input.getAttribute('data-maxlength'),
            div     = document.createElement('div'),
            counter = document.createElement('div'),
            end     = $(input.parentNode).last(),
            width   = $(input).css('width');

        //$(div).css('width', width);

        div.className = 'charCounter';
        $(div).insertAfter(end);
        div.appendChild(counter);
        $(counter).html(input.value.length + '/' + limit);
    },

    updateCharCount: function (e) {
        let limit = this.getAttribute('data-maxlength'),
            input = e.target,
            div, counter;

        div = $('.charCounter');

        counter = $(div).first();
        $(counter).html(input.value.length + '/' + limit);

        if (input.value.length > limit) {
            iw_form.markInvalid(input, '##LG_INPUT_VALUE_OUT_OF_RANGE##');
        } else {
            iw_form.clearStatus(input);
        }
    },

    /*
     * Clears the input status
     */
    _clearStatus: function (e) {
        iw_form.clearStatus(this);
    },

    /*
     * Determines if an input is required or not
     *
     * @return boolean
     */
    isRequired: function (input) {
        return $(input).hasClass('iw-form-req');
    },

    /*
     * Gets the input status image from a particular form input
     *
     * @param HTMLElement el The input next to the status image to get
     */
    getStatusImage: function (el) {
        let img = document.getElementById($(el).attr('id') + '_status_span'),
            div,
            stats;

        if (img) {
            return img;
        }

        div = $(el).closest('.iw-form-input');
        stats = $('.iw-input-status', div);

        if (!stats[0]) {
            return false;
        }

        return stats[0];
    },

    /*
     * Gets the row that a form input is contained in
     *
     * @param HTMLElement el The input in the row to get
     */
    getTableRow: function (el) {
        let tr = $(el).closest('.iw-tr');

        if (!tr) {
            return false;
        }

        return tr;
    },

    getTableCells: function (el) {
        let tr = iw_form.getTableRow(el),
            cells;

        if (tr === false) {
            return false;
        }

        return $(tr).children('td');
    },

    /*
     * Formats an array or string into a string suitable for display in the errors
     *
     * @param string|array
     * @return string
     */
    formatError: function (arrayOrString) {
        if ($.isArray(arrayOrString)) {
            return arrayOrString.join('<br>');
        }
        return arrayOrString;
    },

    /*
     * Marks a given input as invalid (red X)
     *
     * @param HTMLElement input
     * @param string      msg   error message for tooltip
     */
    markInvalid: function (input, msg) {
        this._updateStatus(input, IW.icons.NEGATIVE, this.formatError(msg));
    },

    /*
     * Marks a given input as valid (green check)
     *
     * @param HTMLElement input
     */
    markValid: function (input) {
        this._updateStatus(input, IW.icons.POSITIVE);
    },

    /*
     * Marks a given input as pending (spinner)
     *
     * @param HTMLElement Input
     */
    markPending: function (input) {
        this._updateStatus(input, IW.icons.PENDING);
    },

    markMessages: function (input, msg) {
        this._updateStatus(input, IW.icons.INFO, this.formatError(msg));
    },

    /*
     * Removes all status from the given input
     *
     * @param HTMLElement input
     */
    clearStatus: function (input) {
        let error = document.getElementById('iw-error-' + input.id);
        if ($(input).data('iw-prevent-xhr-validation')) {
            return;
        }
        this._updateStatus(input, IW.icons.BLANK);
        if (error) {
            $(error).removeClass('iw-error-populated');
        }
    },

    _updateStatus: function (input, css_class, msg) {
        msg = msg || '';
        let stat  = iw_form.getStatusImage(input),
            error = $('#iw-error-' + input.id);

        if (stat) {
            iw_lib.updateIWIcon(stat, css_class);
            if (css_class !== IW.icons.BLANK) {
                $(stat).attr('data-original-title', msg);
                $(stat).tooltip();
            } else {
                $(stat).tooltip('hide');
            }

            if (error) {
                error.innerHTML = msg;
                error.addClass('iw-error-populated');
            }
        }//end if
    },

    /*
     * Checks if an input has an error
     *
     * @param HTMLElement input
     */
    hasError: function (input) {
        let stat = iw_form.getStatusImage(input);
        if (stat === false) {
            return false;
        }
        return $(stat).hasClass(IW.icons.NEGATIVE) ||
               $(stat).hasClass(IW.icons.ALERT);
    },

    enableInputs: function (form) {
        let i, el_count = form.elements.length;
        for (i = 0; i < el_count; ++i) {
            if (form.elements[i].type && form.elements[i].type === 'checkbox') {
                continue;
            }
            form.elements[i].disabled = false;
        }
    },

    onSubmit: function (e) {
        let form      = e.target,
            form_name = $(form).attr('name');

        iw_form.saveRichTextEditors();
        if (iw_form.validateFormForErrors(e) === true) {
            // First, enable all the form fields, in case they were disabled
            iw_form.enableInputs(form);
            $(form).trigger(form_name + 'Submit', e);
        } else {
            e.stopPropagation();
            e.preventDefault();
        }
    },

    saveRichTextEditors: function () {
        let editor;

        for (editor in iw_form.richTextEditors) {
            if (iw_form.richTextEditors.hasOwnProperty(editor)) {
                iw_form.richTextEditors[editor].saveHTML();
            }
        }
    },

    /*
     * Validates that the form has no errors
     *
     * @param Event e
     */
    validateFormForErrors: function (e) {
        let is_valid = true,
            form     = e.target,
            i, len, current_input, limit, file;

        iw_form.clearStatus(iw_form.getSubmitButton(form));

        for (i = 0, len = form.elements.length; i < len; ++i) {
            current_input = form.elements[i];

            if (current_input.type === 'file') {
                if (current_input.files &&
                    current_input.files[0] &&
                    iw_js_data &&
                    iw_js_data.post_max_size) {
                    file = current_input.files[0];
                    limit = iw_js_data.post_max_size;
                    if (file.size > limit) {
                        e.stopPropagation();
                        e.preventDefault();
                        iw_form.markInvalid(
                            iw_form.getSubmitButton(form),
                            '##LG_UPLOADMAXSIZELIMIT_ALT##'
                        );
                        return false;
                    }
                }
                continue;
            }

            iw_form.checkValidateRequired(current_input);
            if (current_input.name === 'submit' ||
                current_input.name === 'cancel') {
                continue;
            }
            if (iw_form.hasError(current_input) === true) {
                // Flag that the submit should be cancelled
                is_valid = false;
            }
        }//end for

        if (is_valid === false) {
            e.stopPropagation();
            e.preventDefault();
            iw_form.markInvalid(
                iw_form.getSubmitButton(form),
                '##LG_INPUT_FORM_DID_NOT_VALIDATE##'
            );
            return false;
        }

        return true;
    },

    onCancel: function (e) {
        e.stopPropagation();
        e.preventDefault();

        if (iw_form.isInPopup(e.target)) {
            $('#iw-right-modal').modal('hide');
        } else {
            history.back();
        }
    },

    /*
     * Event wrapper to validate a given input
     *
     * @param Event  e
     * @param object form
     */
    validate: function (e) {
        let form = $(e.target).closest('form').get();
        e.stopPropagation();
        e.preventDefault();
        if (iw_form.formSubmitStatusCache[$(form).attr('name')] === false) {
            iw_form.doValidate(e.target);
        }
    },

    /*
     * Checks if further validation is required.
     *
     * Returns true or false. If true, it means that it did nothing, if false it
     * means that it did the validation, so the next step should not continue
     *
     * @param HTMLElement el
     * @return boolean READ DESCRIPTION ABOVE
     */
    checkValidateRequired: function (el) {
        let value,
            style_check1,
            style_check2,
            empty_array = [],
            tr          = $(el).closest('tr');


        if ($(el).hasClass('tt-hint')) {
            // don't want to a validate type hints
            // for autocomplete
            return false;
        }

        // Only check if the value is blank
        if ($(el).val() === '' && $(el).attr('value') === '') {
            style_check1 = $(el).css('display');
            style_check2 = $(tr).css('display');

            if (style_check1 !== false &&
                style_check1 !== 'none' &&
                style_check2 !== false &&
                style_check2 !== false) {
                if ($(el).css('display') !== 'none' &&
                    $(tr).css('display') !== 'none') {
                    if (iw_form.isRequired(el) === true) {
                        iw_form.markInvalid(el, '##LG_INPUT_IS_REQUIRED##');
                        return false;
                    } else {
                        iw_form.markValid(el);
                        return false;
                    }
                }
            }
        } else if ($(el).attr('type') === 'checkbox' &&
                   $(el).hasClass('iw-unlim-check') === false &&
                   $(el).hasClass('iw-check-all') === false) {
            if ($(el).attr('checked')) {
                iw_form.markValid(el);
                return false;
            }
            if (iw_form.isRequired(el) === false) {
                iw_form.markValid(el);
                return false;
            }

            value = iw_form.getValue(el);

            if (value.join(',') !== '') {
                iw_form.markValid(el);
                return false;
            } else {
                iw_form.markInvalid(el, '##LG_INPUT_IS_REQUIRED##');
                return false;
            }
        }//end if
        return true;
    },

    isAutoComplete: el => $(el).hasClass('iw-form-ac'),

    /*
     * Validates an input
     *
     * @param HTMLElement el
     */
    doValidate: function (el) {
        if (iw_form.isAutoComplete(el)) {
            iw_form.clearStatus(el);
            return;
        }

        let div        = $(el).closest('div'),
            form       = $(el).closest('form')[0],
            stat       = iw_form.getStatusImage(el),
            args,
            check,
            element_name,
            radios,
            linked, linkedInput, id;

        if ($(el).css('display') === 'none') {
            // Not displayed, so the user can't do anything about it
            return;
        }

        if ($(el).data('iw-prevent-xhr-validation')) {
            return;
        }

        if ($(el).attr('type') && $(el).attr('type') === 'file') {
            return;
        }

        if ($(el).hasClass('iw-form-radio-custom')) {
            return;
        }

        if ($(el).attr('name') === 'confirm_password') {
            // Stupid. The confirm password needs the input of the actual password
            if ($(el).val() !== form.password.value) {
                iw_form.markInvalid(el, '##LG_PASSWORDSDONOTMATCH##');
                return;
            } else if ($(el).val() !== '') {
                iw_form.markValid(el);
                return;
            }
        }

        element_name = $(el).attr('name');

        radios = $(div).children('input');

        if (radios.length) {
            if (radios[0].type === 'radio') {
                element_name = radios[0].name;
                if ($(el).val() === '') {
                    iw_form.clearStatus(el);
                    return;
                }
            }
        }

        // First check the required status
        if (iw_form.checkValidateRequired(el) === false) {
            return;
        }

        // Now check the status of unlimited inputs
        let current_val = $(el).val();
        if (current_val === '999999999' || current_val === 'Unlimited') {
            check = $('.iw-unlim-check', div);
            if (check[0]) {
                iw_form.markValid(el);
                return;
            }
        }
        // If it's a select, it's always valid unless someone is being a real turd
        // It'll be validated on the server anyway, so return
        if ($(el).is('select')) {
            iw_form.markValid(el);
            return;
        }

        if (iw_js_data.xhr_validate !== '1') {
            return;
        }

        // Clear the status image
        iw_form.markPending(el);

        args = iw_form.getHiddenInputValues(form);

        if (form.target_domain) {
            args.target_domain = $(form.target_domain).val();
        }

        args.form = form.iw_form_class_name.value;
        args.ctrl = form.iw_ctrl_name.value;
        args.input = element_name;
        args.value = $(el).val();

        linked = $(el).data('iw-linked');

        if (linked) {
            linked = linked.split(',');

            for (id in linked) {
                if (linked.hasOwnProperty(id)) {
                    id = linked[id];
                    linkedInput = document.getElementById(id);

                    if (linkedInput) {
                        args[linkedInput.attributes.getNamedItem('name').value] = iw_form.getValue(
                            linkedInput
                        );
                    }
                }
            }
        }

        iw_xhr.getReply(iw_form_validate_handler, args, stat);
    },

    getHiddenInputValues: function (form, include_controller_info) {
        let temp, temp_length, i, output = {},
            controller_info              = {
                'iw_action_name'    : true,
                'action'            : true,
                'iw_ctrl_name'      : true,
                'iw_form_class_name': true,
                'security_token'    : true
        };

        include_controller_info = include_controller_info || false;

        // Get all inputs
        temp        = $('input', form);
        temp_length = temp.length;
        // Loop through and add the values of any hidden inputs to the form
        for (i = 0; i < temp_length; ++i) {
            if (temp[i].type === 'hidden') {
                if (include_controller_info ===
                    false &&
                    controller_info[temp[i].name]) {
                    continue;
                }

                output[temp[i].name] = temp[i].value;
            }
        }

        return output;
    },

    /*
     * Resets a form, but keeps the current values of show/hide flags
     *
     * @param string|form id_or_form
     */
    reset: function (id_or_form) {
        let form                = iw_lib.getEl(id_or_form),
            show_selects        = $('select.iw-show'),
            show_selects_length = show_selects.length,
            toggle_shows        = $('input[data-iw-show]', form),
            toggle_shows_length = toggle_shows.length,
            current_values      = {},
            j;
        for (j = 0; j < toggle_shows_length; j++) {
            current_values[toggle_shows[j].id] = (toggle_shows[j].checked);
        }
        for (j = 0; j < show_selects_length; j++) {
            current_values[show_selects[j].id] = iw_form.getValue(show_selects[j]);
        }
        form.reset();
        for (j = 0; j < toggle_shows_length; j++) {
            toggle_shows[j].checked = current_values[toggle_shows[j].id];
        }
        for (j = 0; j < show_selects_length; j++) {
            iw_form.setValue(show_selects[j], current_values[show_selects[j].id]);
        }
    },

    /*
     * Checks all the checkboxes related to the input
     *
     * "related to the input" means "in the same form cell", which is how the Form
     * class and template will display it automatically.
     *
     * @param event e
     */
    checkAll: function (e) {
        e.stopPropagation();
        e.preventDefault();
        let checks        = iw_form._getCheckboxesForCheckAll(this),
            checks_length = checks.length,
            i;

        for (i = 0; i < checks_length; ++i) {
            checks[i].checked = 'checked';
        }
    },

    /*
     * Gets the checkboxes relevant to a checkall/none event.
     *
     * @param  {element} el Check all or none that was clicked.
     *
     * @return {array}
     */
    _getCheckboxesForCheckAll: function (el) {
        let div    = $(el).closest('.iw-form-input');
        return $('input[type=checkbox]', div);
    },

    /*
     * Unchecks all the checkboxes related to the input
     *
     * "related to the input" means "in the same div", which is how the Form class
     * and template will display it automatically.
     *
     * @param event e
     */
    checkNone: function (e) {
        e.stopPropagation();
        e.preventDefault();
        let checks        = iw_form._getCheckboxesForCheckAll(this),
            checks_length = checks.length,
            i;

        for (i = 0; i < checks_length; ++i) {
            checks[i].checked = '';
        }
    },

    /*
     * Cache for holding properties of animated elements
     */
    inputPropertyCache: {},

    /*
     * Generates strong passwords.
     */
    generatePassword: function (e) {
        let input     = document.getElementById(this.getAttribute('data-iw-input')),
            maxlength = input.getAttribute('maxlength');
        e.preventDefault();
        e.stopPropagation();
        iw_xhr.getReply(iw_form_generate_password_data_handler, maxlength, this);
        iw_form.markPending(input);
    },

    /*
     * Method for scoring password quality.
     */
    passwordStrength: function (pwd) {
        let score    = 0,
            has_num  = false,
            has_char = false,
            result;

        if (pwd.length < 1) {
            return "";
        }
        if (pwd.length < 4) {
            return "##LG_VERY_WEAK_PASSWORD##";
        }
        if (pwd.length >= 7) {
            score++;
        }
        if (pwd.length >= 12) {
            score++;
        }
        if (pwd.match(/[a-z]/) && pwd.match(/[A-Z]/)) {
            score++;
        }
        if (pwd.match(/[0-9]/)) {
            has_num = true;
        }
        if (pwd.match(/.[!,@,#,$,%,\^,&,*,?,_,~,-,£,(,)]/)) {
            has_char = true;
        }

        if (score === 0) {
            result = '##LG_VERY_WEAK_PASSWORD##';
        } else if (score === 1) {
            result = '##LG_WEAK_PASSWORD##';
        } else if (score === 2 && has_num) {
            result = '##LG_STRONG_PASSWORD##';
        } else if (score === 3 && (has_num || has_char)) {
            result = '##LG_VERY_STRONG_PASSWORD##';
        } else {
            result = '##LG_MEDIUM_PASSWORD##';
        }

        return result;
    },

    // TIMDO: Maybe a progress bar along the bottom of the input?
    updateStrengthMessage: function (eventName, args) {
        let reply = args.reply,
            el    = args.el,
            id    = ($(el).attr('id')).replace(/_status_span/g, ""),
            strength_id = id + '_strength',
            input       = document.getElementById(id);
        if (reply.getReplyCode() !== 204) {
            return;
        }

        if ($(input).hasClass('in-password') === false) {
            return;
        }
        if (reply.getReplyData()[0] === '##LG_COMMON_PASSWORD##') {
            document.getElementById(strength_id).innerHTML = reply.getReplyData()[0];
            document.getElementById(strength_id).className = 'iw-common';
        }
    },

    /*
     * Provides keydown support for as-you-type password rating.
     */
    keydownPasswordStrength: function (e) {
        let el  = e.target,
            pwd = $(el).val();
        iw_form.processPasswordStrength(el, pwd);
    },

    /*
     * Inserts/updates password strength element.
     */
    processPasswordStrength: function (el, pwd) {
        let strength    = iw_form.passwordStrength(pwd),
            strength_id = $(el).attr('id') + '_strength',
            el_idtag    = $(el).attr('id').substring($(el).attr('id').indexOf('_')),
            name_id,
            name,
            email_id,
            email,
            classes     = {
                "##LG_VERY_WEAK_PASSWORD##"  : "iw-veryweak",
                "##LG_WEAK_PASSWORD##"       : "iw-weak",
                "##LG_MEDIUM_PASSWORD##"     : "iw-medium",
                "##LG_STRONG_PASSWORD##"     : "iw-strong",
                "##LG_VERY_STRONG_PASSWORD##": "iw-verystrong",
                "##LG_COMMON_PASSWORD##"     : "iw-common"
        };

        name_id = 'nickname' + el_idtag;
        email_id = 'email' + el_idtag;

        if (document.getElementById(name_id) && document.getElementById(email_id)) {
            name = document.getElementById(name_id).value;
            email = document.getElementById(email_id).value;

            if (name === pwd ||
                email === pwd ||
                email.substring(0, email.indexOf('@')) === pwd) {
                strength = "##LG_VERY_WEAK_PASSWORD##";
            }
        }

        if (document.getElementById(strength_id)) {
            document.getElementById(strength_id).innerHTML = strength;
            document.getElementById(strength_id).className = classes[strength];
            document.getElementById(strength_id).innerHTML = strength;
            document.getElementById(strength_id).className = classes[strength];
        }
    },

    /*
     * When unlimited is checked, fills in the correct value for the input
     *
     * @param event e
     */
    unlimitedCheck: function (e) {
        e.preventDefault();
        e.stopPropagation();
        iw_form.doUnlimitedCheck(e.target);
    },

    /**
     * set input field to unlimited & clear error if any
     */
    doUnlimitedCheck: function (unlimited_button) {
        let $input = $(unlimited_button).closest('.input-group').find('input');
        $input.val('##LG_UNLIMITED##');
        iw_form.markValid($input.get()[0]);
    },

    getAction: function (form) {
        let f = $(form);
        return this.getValue(f.get()[0].iw_action_name);
    },

    getCtrl: function (form) {
        let f = $(form);
        return this.getValue(f.get()[0].iw_ctrl_name);
    },

    /*
     * This should ultimately be expanded to handle multi-selects, checkboxes, etc
     */
    getValue: function (input) {
        let checks, iw_id, esc_iw_id, output = [], i, check_length,
            namePrefix, form, inputs, inputs_length;

        input = iw_lib.getEl(input);

        if (!input) {
            return '';
        }

        if (!input.tagName) {
            return '';
        }

        if (input.tagName.toLowerCase() === 'input') {
            if (input.type === 'checkbox') {
                iw_id = $(input).attr('data-iw-id');
                // escape the iw_id for use in the selectory query below
                esc_iw_id = iw_id.replace('[', '\\[').replace(']', '\\]');
                checks = $('input[data-iw-id=' + esc_iw_id + ']');
                check_length = checks.length;

                for (i = 0; i < check_length; ++i) {
                    if (checks[i].checked) {
                        output.push(checks[i].value);
                    }
                }
                return output;
            } else if (input.type === 'hidden') {
                if (input.name.indexOf('[') === -1) {
                    return input.value;
                } else {
                    namePrefix = input.name.substring(0, input.name.indexOf('['));
                    namePrefix = namePrefix + '\\[';
                    form = $(input).closest('form');
                    inputs = $(form.elements).filter('input[name^=' + namePrefix + ']');
                    inputs_length = inputs.length;

                    for (i = 0; i < inputs_length; ++i) {
                        output.push(inputs[i].value);
                    }
                    return output;
                }
            }//end if
            return input.value;
        }//end if

        if (input.tagName.toLowerCase() === 'select') {
            return input.options[input.selectedIndex].value;
        }
    },

    setValue: function (input, value) {
        let check;
        if (input.hasOwnProperty(0) &&
            input[0].type &&
            input[0].type === 'checkbox') {
            // Then it's a checkbox.
            if ($.isArray(value) === false) {
                value = [value];
            }

            for (check in input) {
                if (input.hasOwnProperty(check)) {
                    check = input[check];

                    if (value.indexOf(check.value) !== -1) {
                        check.checked = 1;
                    } else {
                        check.checked = 0;
                    }
                }
            }
            return;
        }

        input = iw_lib.getEl(input);

        if (input.tagName.toLowerCase() === 'input') {
            input.value = value;
            return;
        }

        if (input.tagName.toLowerCase() === 'select') {
            input.value = value;
            return;
        }
    },

    setFocus: function (form, name) {
        form = iw_lib.getEl(form);
        form.elements[name].focus();
    },

    /*
     * Enables form reset when subscribed to success events.
     */
    enableResetOption: function (e, args) {
        let options = args.options;

        options.prevent_reset = false;
    },

    /*
     * Prevents form reset when subscribed to success events.
     *
     * This is the default.
     *
     * @param string eventType YUI Custom event name
     * @param array  args      args[0] = the reply object, args[1] = options array
     */
    disableResetOption: function (e, args) {
        let options = args.options;

        options.prevent_reset = true;
    },

    enableSubmit: function (form) {
        /*
         * Only enable the button and hide spinner if the button is really there
         */
        if (form) {
            if (iw_form.getSubmitButton(form) && iw_form.getSubmitButton(form).disabled) {
                iw_form.getSubmitButton(form).disabled = 0;
                // Clear the spinner
                if ($(iw_form.getStatusImage(iw_form.getSubmitButton(form))).hasClass(IW.icons.PENDING)) {
                    iw_form.clearStatus(iw_form.getSubmitButton(form));
                }
            }
        }
    }
};

/*
 * Handles requests back to the server for line-by-line validation
 */
iw_form_validate_handler = {
    responder: 'Form_InputValidator',

    /*
     * Ok result
     */
    handle200: function (reply, el) {
        iw_form.markValid(el);
        $(el).prevAll('.iw-in').trigger('validateCompleteEvent', {reply, el});
    },

    handle201: function (reply, el) {
        iw_form.markMessages(el, reply.getReplyData());
        $(el).prevAll('.iw-in').trigger('validateCompleteEvent', {reply, el});
    },

    /*
     * Not Ok result
     */
    handle204: function (reply, el) {
        iw_form.markInvalid(el, reply.getReplyData().join('<br>'));
        $(el).prevAll('.iw-in').trigger('validateCompleteEvent', {reply, el});
    },

    handle403: function (reply, el) {
        iw_form.markInvalid(el, reply.getReplyText());
    }
};

iw_form_autocomplete_data_handler = {
    responder: 'Form_AutoCompleteData',

    handle201: function (reply, el) {
        IW_AC_DATA[$(el).attr('id')] = $.grep(reply.getReplyData(), function (n) {
            return (n.v != null);
        });

        iw_form.initAutoCompleteInput(el);
    }
};

iw_form_label_handler = {
    responder: 'Form_LabelRefresh',

    handle201: function (reply, form) {
        let old_label = $('.iw-form-label', form);

        old_label[0].innerHTML = reply.getReplyData();
    }
};

iw_form_generate_password_data_handler = {
    responder: 'GeneratePassword',

    handle201: function (reply, generate_button) {
        let password_input_id = $(generate_button).data('iw-input');
        let confirm_id = password_input_id.replace('iw-in-', 'iw-in-confirm_');
        let password = reply.getReplyData();

        let input_password = document.getElementById(password_input_id);
        let input_confirm_password = document.getElementById(confirm_id);

        iw_form.processPasswordStrength(input_password, password);
        iw_form.clearStatus(input_password);
        iw_form.clearStatus(input_confirm_password);

        input_password.value = password;
        input_confirm_password.value = password;

        let $password_wrapper = $(input_password).closest('.input-group').parent();
        iw_form_generate_password_data_handler.addPasswordPreviewDiv($password_wrapper, password);

        // trigger event - necessary for sync password option in sw user edit form.

        $(input_password).trigger('input');
    },

    addPasswordPreviewDiv: function ($group, password) {
        let psw_input_id = $group.first('input').attr('id');
        let id = 'iw-generated-pw-' + psw_input_id;
        let div = `<div id="${id}" class="mt-1 ml-2">${password}</div>`;

        if ($(`#${id}`).length) {
            $(`#${id}`).replaceWith(div);
        } else {
            $group.append(div);
        }
    }
};

/*
 * Special function that adds a listener to a given eventname on the specified
 * element, and on that action submit/routes that form xhr style through the
 * json controller mechanism
 *
 * @param HTMLElement form      The form that will ultimately be submitted
 * @param HTMLElement element   The element that the specified event is listen on
 *                              If null, form is assumed as the element
 * @param string      eventname The event that is listened for: 'change' 'submit' etc
 */
var iw_form_handler = function (form, element, eventname) {
    let formName;
    element = element || form;
    eventname = eventname || 'submit';

    /*
     * The handler callback, this function shows a banner if the reply contains data
     * and pops up an error if success was false.  Also routes to callback, if it was
     * specified when the iw_form_handler was made.  Finally, it  re-enables the
     * submit button if applicable
     *
     * @param iw_xhr.reply r
     * @param boolean      success
     */
    function complete(r, success)
    {
        let redirect,
            i,
            error,
            input,
            d        = r.getReplyData(),
            formName = $(form).attr('name'),
            // Need to re-get the form, because sometimes it's been innerHTMLed and is gone
            newForm  = document.forms[formName],
            ctrl     = iw_form.getCtrl(newForm),
            action   = iw_form.getAction(newForm),
            options  = {
                display_banner  : true,
                display_errors  : true,
                prevent_redirect: false,
                prevent_reset   : true,
                prevent_hide    : false,
        };

        if (success) {
            $(document).trigger('success_' + ctrl + '_' + action, {reply: r, options: options});

            // show success banner
            if (d) {
                if (d.redirect) {
                    if (d.force_redirect ||
                        (window.location.pathname + window.location.search) !==
                        d.redirect &&
                        options.prevent_redirect !==
                        true) {
                        d.message += ' -  ##LG_PLEASEWAIT##';
                    }
                }
                if (options.display_banner) {
                    iw_lib.iw_message_banner(d.message, 'iw-positive');
                }
                if (options.prevent_reset === false) {
                    iw_form.reset(newForm);
                }
            }
        } else {
            iw_form.formSubmitStatusCache[formName] = false;

            if (iw_xhr.events.failure[ctrl] && iw_xhr.events.failure[ctrl][action]) {
                $(document).trigger('failure_' + ctrl + '_' + action, {reply: r, options: options});
                // iw_xhr.events.failure[ ctrl ][ action ].fire( r, options );
            }

            // show error banner
            if (d && options.display_banner) {
                iw_lib.iw_message_banner(d.message, 'iw-negative');
            }

            /*
             * If we got input_errors back, then we should try to mark the inputs
             * in the form as invalid
             */
            if (options.display_errors && d.input_errors) {
                for (i in d.input_errors) {
                    if (d.input_errors.hasOwnProperty(i)) {
                        error = d.input_errors[i];
                        input = form[error.name];
                        // workaround for input_sets.
                        if (!input) {
                            input = form[error.name + '_0'];
                        }
                        // Last resort
                        if (!input) {
                            input = iw_form.getSubmitButton(form);
                            error.err = error.name + ': ' + error.err;
                        }

                        // Hidden inputs don't have the necessary structure around them.
                        if (input.type !== 'hidden') {
                            iw_form.markInvalid(input, error.err);
                        }
                    }
                }
            }//end if
        }//end if

        iw_form.enableSubmit(form);
        toggleDisabledToolbarActions();

        /*
         * If redirect data was part of the response, redirect to the new location
         * if not already there
         */
        if (options.prevent_redirect === false && d) {
            if (d.redirect) {
                redirect = d.redirect;

                if (form.iw_force_redirect) {
                    redirect = form.iw_force_redirect.value;
                    d.force_redirect = true;
                }

                if (d.force_redirect ||
                    (window.location.pathname + window.location.search) !==
                    redirect) {
                    window.setTimeout(function () {
                        window.location = redirect;
                    }, iw_lib.redirect_timeout);
                }
            }
        }
    }

    /*
     * This is the function that actually executes the iw_ctrl_handler / submits the form
     * It spins up the spinner and disables the submit button if applicable
     *
     * @param string type The type of event fired
     * @param array args
     */
    function route(e, args)
    {
        let type = e.type,
            form = null, form_name;

        // if toolbar action, get action name from target element.
        let action = $(e.target).hasClass('btn-toolbar-action') &&  e.target.name;


        // If this is called by a custom event, it's different
        e =  args || e;

        if (e.type === 'submit') {
            // if it's a submit event, then the target IS the form
            form = form || e.target;
        } else {
            // if it's not a submit event, then the target is a form
            // element, and target.form is the form.
            form = e.target.form;
        }

        form_name = $(form).attr('name');

        if (e.type === 'submit') {
            iw_form.formSubmitStatusCache[form_name] = true;
        }

        // stop the event, we're taking over
        e.stopPropagation();
        e.preventDefault();

        // in IE form.submit is defined even if there is no
        // "submit" button in the form, so checking both submit and
        // submit.disabled here
        if (iw_form.getSubmitButton(form) && !iw_form.getSubmitButton(form).disabled) {
            // disable the submit button
            iw_form.getSubmitButton(form).disabled = 1;
            // Show the spinner on the submit button
            iw_form.markPending(iw_form.getSubmitButton(form));
        } else {
            // Show the spinner on the input element that triggered
            // the event
            iw_form.markPending(e.target);
        }

        toggleDisabledToolbarActions(form);


        // fire off to the xhr request
        iw_ctrl_handler.execute(
            form,
            complete,
            iw_form.getCtrl(form),
            action || iw_form.getAction(form)
        );
    }//end function route


    /**
     * toggle prop "disabled" on toolbar actions
     */
    const toggleDisabledToolbarActions = () => {
        $('.btn-toolbar-action').each((index, btn) => {
            let isDisabled = $(btn).prop('disabled');
            $(btn).prop('disabled', !isDisabled);
        });
    }

    // Only subscribe the event if the form is not null
    if (element) {
        // Interrupt the normal subscription and subscribe to the custom event instead
        if (eventname === 'submit') {
            formName = $(form).attr('name');
            $(element).on(formName + 'Submit', element, route);
        } else {
            if (element.id) {
                let selector = `#${element.id}`;
                $(document).off(eventname, selector, route);
                $(document).on(eventname, selector, route);
            } else {
                // default to assign event directly on element
                // if element does not have an id. The preferred
                // way is on the document, so that elements added
                // after initial load also gets the event listeners.
                $(element).on(eventname, element, route);
            }
        }
    }
};

iw_form.init();
